package k.ld34;

import haxe.Timer;
import openfl.Lib;
import openfl.display.Bitmap;
import openfl.display.BitmapData;
import openfl.display.Sprite;
import openfl.filters.DropShadowFilter;
import openfl.geom.Point;
import openfl.geom.Rectangle;
import openfl.text.TextField;
import openfl.text.TextFieldAutoSize;
import openfl.text.TextFormat;
import openfl.text.TextFormatAlign;

import k.lib.graphics.Assets;
import k.lib.graphics.DisplayObject;
import k.lib.graphics.RenderEngine;
import k.lib.graphics.Text;
import k.lib.particles.ParticleEngine;
import k.lib.sound.SoundManager;
import k.lib.utils.LevelEditor;
import k.lib.utils.Utils;
import k.lib.utils.APathEngine;

import k.ld34.entities.AsteroidBelt;
import k.ld34.entities.Entity;
import k.ld34.entities.Player;
import k.ld34.entities.Ship;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

class Loader extends Sprite {
	private var render:RenderEngine;
	private var particles_render:RenderEngine;
	private var particles_engine:ParticleEngine;

	private var callback:Void->Void;
	private var text:TextField;
	private var _width:Int;
	private var renderStep:Int;
	private var timer:Timer;
	private var rendering:Bool;

	private var _stage:Stage;
	private var titleScreen:Bitmap;

	public function new(stage:Stage, width:Int, height:Int) {
		super();
		_stage = stage;
		_width = width;

		// Draw background
		var background_asset:BitmapData = Assets.getBitmapData("assets/background.png");
		graphics.beginBitmapFill(background_asset);
		graphics.drawRect(0,0,width,height);
		graphics.endFill();
		cacheAsBitmap = true;

		// Add textfield
		text = new TextField();
		text.selectable = false;
		text.autoSize = TextFieldAutoSize.NONE;
		text.embedFonts = true;
		text.defaultTextFormat = new TextFormat(Assets.getFontName("visitor1"), 20, 0xCCCCCC, false, false, TextFormatAlign.LEFT);
		text.text = "";
		text.y = height/2;
		text.filters = [new DropShadowFilter(0, 0, 0x0, 1, 6, 6, 2, 1)];
		addChild(text);

		init();
	}

	public function init():Void {
		Entity.resetAll();
		rendering = false;

		_stage.back.clear();
		_stage.safe.clear();
		_stage.enemies.clear();
		_stage.bullets.clear();
		_stage.player.clear();
		_stage.front.clear();
	}

	public function preRender(render:RenderEngine, particles_render:RenderEngine, callback:Void->Void):Void {
		if (timer != null) {
			timer.stop();
			rendering = false;
		}

		this.render = render;
		this.particles_render = particles_render;
		this.callback = callback;

		renderStep = 0;
		timer = new Timer(5);
		timer.run = renderDoStep;
	}

	private function renderDoStep():Void {
		if (rendering || render == null || callback == null) return;
		rendering = true;

		switch (renderStep) {
			case 0:
			update("Pre-rendering sound");
			// GameConsole.log("Sound pre-rendering done");

			case 1:
			update("Pre-rendering graphics");
			// TODO: import aseprite animations
			render.add(Assets.getBitmapData("assets/background.png"), "background");
			render.add(Assets.getBitmapData("assets/Power-Up.png"), "powerup");
			render.add(Assets.getBitmapData("assets/Game-Over.png"), "gameover");
			render.add(Assets.getBitmapData("assets/Title.png"), "title");
			render.add(Assets.getBitmapData("assets/Asteroids2.png"), "asteroid");
			render.add(Assets.getBitmapData("assets/Ship.png"), "ship");
			render.add(Assets.getBitmapData("assets/Player-Maxi-Bullet2.png"), "maxi-bullet");
			render.add(Assets.getBitmapData("assets/Ship-Bullets.png"), "foes-bullet");
			render.add(Assets.getBitmapData("assets/Ship-Bullets-Circle.png"), "foes-bullet-circle");
			render.add(Assets.getBitmapData("assets/Ship-Bullets-V.png"), "foes-bullet-vertical");
			render.add(Assets.getBitmapData("assets/Ship-Shield.png"), "ship-shield");

			var tilesheets:Array<BitmapData> = [
				Assets.getBitmapData("assets/Player-Ship-A.png"),
				Assets.getBitmapData("assets/Player-Ship-B.png")
			];
			var tilename:Array<String> = [
				"front", "", "mid", "safe",
				"", "", "", "", "", "", "", "", // Dirt animation
				"", "front-left", "", "front-right",
				"", "mid-left", "", "mid-right",
				"", "safe-left", "", "safe-right"
			];
			var tileWidth = 32;
			var tileHeight = 32;
			for (x in 0...tilename.length) {
				if (tilename[x] == "") continue;

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[0], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "player-a-" + tilename[x]);

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[1], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "player-b-" + tilename[x]);
			}

			var tilesheets:Array<BitmapData> = [
				Assets.getBitmapData("assets/Ship-A.png"),
				Assets.getBitmapData("assets/Ship-D.png")
			];
			var tilename:Array<String> = [
				"base", "hit", "left", "right"
			];
			var tileWidth = 32;
			var tileHeight = 32;
			for (x in 0...tilename.length) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[0], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "ship-a-" + tilename[x]);

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[1], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "ship-d-" + tilename[x]);
			}

			var tilesheets:Array<BitmapData> = [
				Assets.getBitmapData("assets/Ship-B.png")
			];
			var tilename:Array<String> = [
				"h-base", "h-hit",
				"b-base", "b-hit"
			];
			var tileWidth = 32;
			var tileHeight = 32;
			for (x in 0...tilename.length) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[0], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "ship-" + tilename[x]);
			}

			var tilesheets:Array<BitmapData> = [
				Assets.getBitmapData("assets/Ship-C.png")
			];
			var tilename:Array<String> = [
				"g-base", "g-hit",
				"c-base", "c-hit"
			];
			var tileWidth = 32;
			var tileHeight = 32;
			for (x in 0...tilename.length) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[0], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "ship-" + tilename[x]);
			}

			var tilesheets:Array<BitmapData> = [
				Assets.getBitmapData("assets/Ship-Shield.png")
			];
			var tilename:Array<String> = [
				"red-base", "red-hit",
				"blue-base", "blue-hit"
			];
			var tileWidth = 16;
			var tileHeight = 16;
			for (x in 0...tilename.length) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(tilesheets[0], new Rectangle(x*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "ship-shield-" + tilename[x]);
			}

			var healthbars:BitmapData = Assets.getBitmapData("assets/Player-Health.png");
			var tileWidth = 12;
			var tileHeight = 4;
			for (i in 0...10) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(healthbars, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "healthbar-" + (i + 1));
			}

			var playerShot:BitmapData = Assets.getBitmapData("assets/Player-Fire.png");
			var tileWidth = 32;
			var tileHeight = 32;
			for (i in 0...2) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(playerShot, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "shot-player-" + i);
			}

			var explosion:BitmapData = Assets.getBitmapData("assets/Explosion-Small2.png");
			var tileWidth = 12;
			var tileHeight = 12;
			for (i in 0...4) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(explosion, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "explosion-" + (i+1));
			}

			var star:BitmapData = Assets.getBitmapData("assets/Star.png");
			var tileWidth = 12;
			var tileHeight = 12;
			for (i in 0...5) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(star, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "star-" + (i+1));
			}

			var playerEngine:BitmapData = Assets.getBitmapData("assets/Player-Engine.png");
			var tileWidth = 32;
			var tileHeight = 32;
			for (i in 0...4) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(playerEngine, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "player-engine-" + (i+1));
			}
			Entity.createAnimation("player-engine", [
				"player-engine-1", "player-engine-1", "player-engine-1",
				"player-engine-2", "player-engine-2", "player-engine-2",
				"player-engine-3", "player-engine-3", "player-engine-3",
			]);

			var playerEngineSmoke:BitmapData = Assets.getBitmapData("assets/Player-Smoke.png");
			var tileWidth = 32;
			var tileHeight = 32;
			for (i in 0...8) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(playerEngineSmoke, new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "player-smoke-" + (i+1));
			}
			Entity.createAnimation("player-smoke", [
				"player-smoke-1",
				"player-smoke-2",
				"player-smoke-3",
				"player-smoke-4",
				"player-smoke-5",
				"player-smoke-6",
				"player-smoke-7",
				"player-smoke-8"
			]);

			var bullets:BitmapData = Assets.getBitmapData("assets/Player-Bullets.png");
			var tileWidth = 8; var tileHeight = 8;
			var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
			bd.copyPixels(bullets, new Rectangle(0, 0, tileWidth, tileHeight), Utils.ORIGIN);
			render.add(bd, "player-bullet-a");
			var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
			bd.copyPixels(bullets, new Rectangle(1*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
			render.add(bd, "player-bullet-b");


			var shipShots:Array<BitmapData> = [
				Assets.getBitmapData("assets/Ship-A-Shoot.png"),
				Assets.getBitmapData("assets/Ship-B-Shoot.png"),
				Assets.getBitmapData("assets/Ship-C-Shoot.png"),
				Assets.getBitmapData("assets/Ship-D-Shoot.png")
			];
			var tileWidth = 32;
			var tileHeight = 32;
			for (i in 0...2) {
				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(shipShots[0], new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "shot-ship-a-" + i);

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(shipShots[1], new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "shot-ship-b-" + i);

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(shipShots[2], new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "shot-ship-c-" + i);

				var bd:BitmapData = new BitmapData(tileWidth,tileHeight,true,0x00000000);
				bd.copyPixels(shipShots[3], new Rectangle(i*tileWidth, 0, tileWidth, tileHeight), Utils.ORIGIN);
				render.add(bd, "shot-ship-d-" + i);
			}
			// GameConsole.log("Graphics pre-rendering done");

			case 2:
			update("Pre-rendering fonts");
			Text.prepareFont(render, "VisitorScore", "visitor1", 10, 0xFFFFFF);
			// GameConsole.log("Font pre-rendering done");

			case 3:
			update("Preparing game");
			Entity.defineGroup("asteroid");
			Entity.defineGroup("stars");
			Entity.defineGroup("effects");
			Entity.defineGroup("player-misc");
			Entity.defineGroup("player-bullet", ["foes"]);
			Entity.defineGroup("foes-bullet", ["player"]);
			Entity.defineGroup("foes", ["player", "player-bullet"]);
			Entity.defineGroup("player", ["foes", "foes-bullet"]);

			var back:DisplayObject = new DisplayObject("background");
			_stage.back.addChild(back);

			Main.shipA = new Player(_stage, SHIP_A);
			Main.shipB = new Player(_stage, SHIP_B);
			Main.waveGen = new WaveGenerator(_stage);
			Main.asteroidBelt = new AsteroidBelt(_stage);
			// GameConsole.log("Game ready");

			case 4:
			update("Packing rendered particles");
			particles_render.pack();
			// GameConsole.log("Particles packing done");

			case 5:
			update("Packing rendered graphics");
			render.pack();
			// GameConsole.log("Packing done");

			case 6:
			var bd:BitmapData = Assets.getBitmapData("assets/Title.png");
			titleScreen = new Bitmap(bd);
			// titleScreen.scaleX = Main.GLOBAL_SCALE;
			// titleScreen.scaleY = Main.GLOBAL_SCALE;
			titleScreen.pixelSnapping = NEVER;
			Lib.current.addChild(titleScreen);

			case _:
			if (Main.isDown(Main.BUTTON_A) || Main.isDown(Main.BUTTON_B)) {
				Lib.current.removeChild(titleScreen);
				timer.stop();
				rendering = false;
				callback();
				return;
			}
		}

		renderStep++;
		rendering = false;
	}

	private function update(s:String):Void {
		text.text = s;
		text.width = _width;
		text.x = _width/2 - text.textWidth/2;
	}

}
