package k.ld34;

import openfl.Lib;
import openfl.display.Bitmap;
import openfl.display.BitmapData;
import openfl.display.Shape;
import openfl.events.Event;
import openfl.geom.Matrix;
import openfl.ui.Keyboard;

import k.lib.graphics.Shake;
import k.lib.graphics.Assets;
import k.lib.graphics.DisplayObject;
import k.lib.graphics.DisplayObject.Rect;
import k.lib.graphics.RenderEngine;
import k.lib.graphics.Text;
import k.lib.particles.ParticleEngine;
import k.lib.random.Mersenne;
import k.lib.sound.SoundManager;
import k.lib.utils.Ease;

#if editor
import k.lib.utils.LevelEditor;
#end

#if DEBUG
import haxe.Timer;
import openfl.system.System;
import openfl.text.TextField;

import k.lib.gconsole.GameConsole;
#end

import k.ld34.effects.Star;
import k.ld34.effects.PowerUp;
import k.ld34.entities.Entity;
import k.ld34.entities.Player;
import k.ld34.entities.AsteroidBelt;

class Main extends Shape {
	public static inline var GLOBAL_SCALE:Float = 3.0;
	public static inline var GAME_WIDTH:Int = 120;
	public static inline var GAME_HEIGHT:Int = 160;
	public static inline var MAX_STARS:Int = 5;

	private static var kcodes = new Array<Null<Int>>();
	public static var kframe:Int;

	public static var render:RenderEngine;
	private var particles_render:RenderEngine;

	private var canvas:Shape;
	public static var _stage:Stage;
	public static var particles_engine:ParticleEngine;

	private var bd:BitmapData;
	private var particles_bd:BitmapData;
	private var bitmap:Bitmap;
	private var particles_bitmap:Bitmap;

	private var loader:Loader;
	private var scoreText:Text;

	private var ready:Bool;
	private var forceupdate:Bool;
	private var controlsEnabled:Bool;
	private var _width:Int;
	private var _height:Int;
	private var cut:Rect;
	private static var lost:Int;


	public static var random:Mersenne;
	public static var score:Int;
	public static var waveGen:WaveGenerator;
	public static var slowMotion:Int;
	public static var powerup:PowerUp;
	public static var hasLost:Bool;

	public static var shipA:Player;
	public static var shipB:Player;
	public static var nships:Int;

	public static var stars:Array<Star>;
	public static var asteroidBelt:AsteroidBelt;

	public static var BUTTON_A:Int = Keyboard.X;
	public static var BUTTON_B:Int = Keyboard.C;

	public static var gameInstance:Main;

	public static function main():Void {
		gameInstance = new Main();
		Lib.current.addChild(gameInstance);
	}

	public function new() {
		super();
		ready = false;
		forceupdate = false;
		controlsEnabled = false;
		hasLost = false;
		kframe = 0;
		score = 0;
		lost = 0;
		slowMotion = 0;
		addEventListener(Event.ADDED_TO_STAGE, added);
	}

	public function added(?d:Dynamic):Void {
        removeEventListener(Event.ADDED_TO_STAGE, added);
		Lib.current.stage.addEventListener(Event.ENTER_FRAME, enter);

		#if DEBUG
		Lib.current.stage.addChild(new openfl.display.FPS(10,10,0xFF0000));
		GameConsole.init();
		#end

		Lib.current.stage.addEventListener(flash.events.KeyboardEvent.KEY_DOWN, onKey.bind(true));
		Lib.current.stage.addEventListener(flash.events.KeyboardEvent.KEY_UP, onKey.bind(false));
		// Lib.current.stage.addEventListener(flash.events.Event.DEACTIVATE, function(_) kcodes = new Array<Null<Int>>());

		init();
	}

	private function loader_ready():Void {
		#if flash
		bd = new BitmapData(_width, _height, true, 0x00FFFFFF);
		bitmap = new Bitmap(bd);
		bitmap.scaleX = Main.GLOBAL_SCALE;
		bitmap.scaleY = Main.GLOBAL_SCALE;
		bitmap.pixelSnapping = NEVER;
		Lib.current.addChild(bitmap);

		particles_bd = new BitmapData(_width, _height, true, 0x00FFFFFF);
		particles_bitmap = new Bitmap(particles_bd);
		particles_bitmap.scaleX = Main.GLOBAL_SCALE;
		particles_bitmap.scaleY = Main.GLOBAL_SCALE;
		particles_bitmap.pixelSnapping = NEVER;
		Lib.current.addChild(particles_bitmap);
		#end

		// Score text
		scoreText = new Text(Std.int(GAME_WIDTH / 2), 5, "0", "VisitorScore", Text.ALIGN_CENTER);
		_stage.asteroid.addChild(scoreText);

		cut = {x: 0, y: 0, w: GAME_WIDTH, h: GAME_HEIGHT};
		stars = [];

		// kcodes[Main.BUTTON_A] = null;
		// kcodes[Main.BUTTON_B] = null;

		ready = true;
		forceupdate = true;
		// controlsEnabled = true;
		if (loader != null && Lib.current.contains(loader)) Lib.current.removeChild(loader);
	}

	private function again():Void {
		// Clean up
		if (canvas != null && Lib.current.contains(canvas)) Lib.current.removeChild(canvas);
		#if flash
		if (bitmap != null && Lib.current.contains(bitmap)) Lib.current.removeChild(bitmap);
		if (particles_bitmap != null && Lib.current.contains(particles_bitmap)) Lib.current.removeChild(particles_bitmap);
		#end

		init();
	}

	private function init():Void {
		random = new Mersenne(Std.int(42 + Date.now().getTime()));
		_width = Std.int(Lib.current.stage.stageWidth);
		_height = Std.int(Lib.current.stage.stageHeight);

		_stage = new Stage();
		_stage.checkNeedsUpdate = function() { return Entity.needsUpdate; };

		canvas = new Shape();
		Lib.current.addChild(canvas);

		// Init controls
		kcodes = new Array<Null<Int>>();

		Entity.GRID_SIZE = 32;

		ready = false;
		forceupdate = false;
		controlsEnabled = false;
		hasLost = false;
		kframe = 0;
		score = 0;
		lost = 0;
		slowMotion = 0;
		nships = 2;

		render = new RenderEngine(_width, _height);
		render.setFlag("alpha", true);
		render.setFlag("rotation", false);
		render.setFlag("rgb", true);

		particles_render = new RenderEngine(_width, _height);
		particles_render.setFlag("alpha", true);
		particles_render.setFlag("scale", true);
		particles_render.setFlag("rgb", true);

		particles_engine = new ParticleEngine();

		// Pre-rendering
		loader = new Loader(_stage, _width, _height);
		Lib.current.addChild(loader);
		loader.preRender(render, particles_render, loader_ready);
	}

	@:access(k.lib.graphics.Entity)
	public inline function enter(event:Event):Void {

		if (ready) {
			if (!controlsEnabled && kframe > 10) controlsEnabled = true;

			if (kframe % 30 == 0 && !hasLost) {
				if (nships == 2 && shipA.state == ACTIVE && shipB.state == ACTIVE) {
					score += waveGen.diff;
				} else if (nships == 2 && (shipA.state == ACTIVE || shipB.state == ACTIVE)) {
					score += Std.int(waveGen.diff / 4);
				}
				scoreText.updateText(parseScore());
			}

			if (slowMotion > 0 && kframe % 2 > 0) {
				kframe++;
				slowMotion--;

				if (slowMotion <= 0) powerup.destroy();
				return;
			}

			// Update stars
			if (kframe % 3 == 0 && stars.length < MAX_STARS) {
				var star:Star = new Star(_stage);
				stars.push(star);
			}

			// Game over animation
			if (lost > 0 && kframe - lost == 60) {
				var go:DisplayObject = new DisplayObject("gameover", DisplayObject.ALIGN_CENTER);
				go.x = Std.int(Main.GAME_WIDTH / 2);
				go.y = Std.int(Main.GAME_HEIGHT / 2);
				_stage.front.addChild(go);
			} else if (lost > 0 && kframe - lost > 120) {
				if (isDown(Main.BUTTON_A) || isDown(Main.BUTTON_B)) {
					ready = false;
					again();
				}
			} else {
				waveGen.update();

				Entity.needsUpdate = forceupdate;
				if (controlsEnabled) {
					shipA.setState(isDown(Main.BUTTON_A) ? PROTECTED : ACTIVE);
					shipB.setState(isDown(Main.BUTTON_B) ? PROTECTED : ACTIVE);
				}

				if (isToggled(Keyboard.M)) SoundManager.toggleMute();
				_stage.prepare();

				#if flash
					if (bd != null) render.renderBitmapData(bd, _stage, cut, true);
					// if (particles_bd != null) particles_render.renderBitmapData(particles_bd, particles_engine, true);
				#else
					render.renderTiles(canvas.graphics, _stage, true, false);
					particles_render.renderTiles(canvas.graphics, particles_engine, false, false);
				#end

				forceupdate = false;
			}

			if (!hasLost) scoreText.updateText(parseScore());
			kframe++;
		}
	}

	public static function gameover():Void {
		hasLost = true;

		if (lost == 0) {
			lost = kframe;
		}
	}

	public function parseScore():String {
		var ret:String = score + "";
		while (ret.length < 6) ret = "0" + ret;
		ret = ret.substr(0, 3) + "." + ret.substr(3, 3);
		return ret;
	}

	private inline function updateData():Bool {
		return true;
	}

	private static function onKey(down:Bool, e:flash.events.KeyboardEvent) {
		event(e.keyCode, down);
	}

	public static function event(code:Int, down:Bool) {
		kcodes[code] = (down) ? kframe : null;

		// release Ctrl when Alt pressed (AltGr generates Ctrl+Alt)
		if (code == 18 && !down) kcodes[17] = null;
	}

	public static inline function isDown(c) {
		return kcodes[c] != null;
	}

	public static inline function isToggled(c) {
		return kcodes[c] == kframe;
	}

}
