package k.ld34;

import k.lib.random.Randomizer;
import k.lib.sound.SoundManager;

import k.ld34.waves.*;
import k.ld34.effects.PowerUp;
import k.ld34.entities.ShipA;
import k.ld34.entities.ShipB;
import k.ld34.entities.ShipC;
import k.ld34.entities.ShipD;
import k.ld34.entities.ShipE;
import k.ld34.entities.ShipF;
import k.ld34.entities.ShipG;
import k.ld34.entities.ShipH;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

enum WaveID {
	A_Snake;
	A_Diag;
	A_Line;
	A_M;
	A_W;
	B;
	C;
	D_Snake;
	D_Line;
	D_M;
	D_W;
	E;
	F;
	G;
	H;
}

typedef WaveProba = {
	wave:WaveID,
	prob:Int
}

class WaveGenerator {
	private var stage:Stage;
	public var diff:Int;
	private var frameToDiffUp:Int;
	private var framesToDiffUp:Array<Int>;
	private var frameToNextWave:Int;
	private var minFramesToNextWave:Int;
	private var maxFramesToNextWave:Int;

	public function new(s:Stage) {
		stage = s;
		diff = 0;
		framesToDiffUp = [
			10*60,
			10*60,
			10*60,
			10*60,
			15*60,
			15*60,
			15*60,
			15*60,
			20*60,
			20*60,
			20*60,
			20*60,
			20*60,
			30*60,
			30*60
		];

		frameToDiffUp = framesToDiffUp[0];
		frameToNextWave = 1;

		// TODO
		minFramesToNextWave = 250;
		maxFramesToNextWave = 400;
	}

	private function powerup(i:Int):Void {
		// Inform ships
		Main.shipA.powerUp(i);
		Main.shipB.powerUp(i);

		// Activate slow motion
		Main.slowMotion = 40;

		// GFX
		Main.powerup = new PowerUp(stage);
		Main.powerup.setCoordinates(Std.int(Main.GAME_WIDTH / 2), 100);
	}


	public function update():Void {
		if (Main.hasLost) return;

		if (frameToDiffUp > 0 && Main.kframe >= frameToDiffUp) {
			diff++;
			Main.score += diff * 100;

			if (diff == 3) powerup(1);
			if (diff == 6) powerup(2);
			if (diff == 10) powerup(3);

			frameToDiffUp = (diff < framesToDiffUp.length) ? Main.kframe + framesToDiffUp[diff] : 0;
			// GameConsole.log("DIFF UP: " + diff);

			// TODO
			minFramesToNextWave -= 6;
			if (minFramesToNextWave < 130) minFramesToNextWave = 130;
			maxFramesToNextWave -= 8;
			if (maxFramesToNextWave < 200) minFramesToNextWave = 200;
		}

		if (Main.kframe >= frameToNextWave) {
			frameToNextWave = Main.kframe + Main.random.intBetween(minFramesToNextWave, maxFramesToNextWave);
			var wave:Null<WaveID> = randomWave();
			if (wave != null) waveFactory(wave);
		}
	}

	private function randomWave():Null<WaveID> {
		var availableWaves:Array<WaveProba> = [];

		switch (diff) {
			case 0:
			availableWaves.push({wave: A_Snake, prob: 10});
			availableWaves.push({wave: A_Line, prob: 5});

			case 1:
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: A_Line, prob: 5});
			availableWaves.push({wave: A_Diag, prob: 5});

			availableWaves.push({wave: B, prob: 10});

			case 2:
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: A_Line, prob: 5});
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: A_M, prob: 5});
			availableWaves.push({wave: A_W, prob: 5});

			availableWaves.push({wave: C, prob: 15});

			case 3:
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: A_M, prob: 5});
			availableWaves.push({wave: A_W, prob: 5});

			availableWaves.push({wave: E, prob: 10});

			case 4:
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: A_M, prob: 5});
			availableWaves.push({wave: A_W, prob: 5});

			availableWaves.push({wave: B, prob: 5});
			availableWaves.push({wave: C, prob: 10});

			case 5:
			availableWaves.push({wave: D_Line, prob: 5});
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: A_M, prob: 5});
			availableWaves.push({wave: A_W, prob: 5});

			availableWaves.push({wave: B, prob: 10});
			availableWaves.push({wave: C, prob: 5});
			availableWaves.push({wave: E, prob: 5});

			case 6:
			availableWaves.push({wave: D_Line, prob: 5});
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: A_Snake, prob: 5});
			availableWaves.push({wave: D_Snake, prob: 5});
			availableWaves.push({wave: A_M, prob: 5});
			availableWaves.push({wave: A_W, prob: 5});

			availableWaves.push({wave: H, prob: 10});
			availableWaves.push({wave: C, prob: 10});
			availableWaves.push({wave: E, prob: 10});

			default:
			availableWaves.push({wave: D_Line, prob: 5});
			availableWaves.push({wave: A_Diag, prob: 5});
			availableWaves.push({wave: D_Snake, prob: 5});
			availableWaves.push({wave: D_M, prob: 5});
			availableWaves.push({wave: D_W, prob: 5});

			availableWaves.push({wave: H, prob: 10});
			availableWaves.push({wave: F, prob: 10});
			availableWaves.push({wave: G, prob: 10});
		}

		var probTotal:Int = 0;
		for (i in 0...availableWaves.length) probTotal += availableWaves[i].prob;

		var randomizer:Randomizer<WaveID> = new Randomizer<WaveID>(Main.random, probTotal);
		for (i in 0...availableWaves.length) {
			randomizer.allocate(availableWaves[i].wave, availableWaves[i].prob);
		}

		return randomizer.random();
	}


	private function waveFactory(waveId:WaveID):Void {
		switch (waveId) {
			case A_Snake:
			new SnakeWave<ShipA>(stage, diff);

			case A_Line:
			new LineWave<ShipA>(stage, diff, Main.random.intBelow(10) > 5 ? -1 : 1);

			case A_Diag:
			if (diff > 5 && Main.random.intBelow(10) > 6) {
				new DiagWave(stage, diff, 1);
				new DiagWave(stage, diff, -1);
			} else {
				new DiagWave(stage, diff, Main.random.intBelow(10) > 5 ? -1 : 1);
			}

			case A_M:
			new MWave<ShipA>(stage, diff);

			case A_W:
			new WWave<ShipA>(stage, diff);

			case B:
			new ShipB(stage, diff, Main.random.intBelow(10) > 5 ? -1 : 1);

			case C:
			new ShipC(stage, diff);

			case D_Snake:
			new SnakeWave<ShipD>(stage, diff);

			case D_Line:
			new LineWave<ShipD>(stage, diff, Main.random.intBelow(10) > 5 ? -1 : 1);

			case D_M:
			new MWave<ShipD>(stage, diff);

			case D_W:
			new WWave<ShipD>(stage, diff);

			case E:
			new ShipE(stage, diff);

			case F:
			new ShipF(stage, diff);

			case G:
			new ShipG(stage, diff);

			case H:
			new ShipH(stage, diff, Main.random.intBelow(10) > 5 ? -1 : 1);
		}
	}
}
