package k.ld34.entities;

import k.lib.graphics.DisplayObject;
import k.lib.utils.KMath;
import k.lib.utils.Ease;
import k.lib.sound.SoundManager;

import k.ld34.effects.Shot;
import k.ld34.effects.Engine;
import k.ld34.effects.EngineSmoking;
import k.ld34.effects.Explosion;
import k.ld34.entities.Entity;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

enum PlayerShip {
	SHIP_A;
	SHIP_B;
}

enum PlayerState {
	ACTIVE;
	PROTECTED;
}

enum MovementPhase {
	PHASE_1;
	PHASE_2;
	PHASE_3;
	PHASE_4;
}

class Player extends Entity {
	public var shipId:PlayerShip;
	public var state:PlayerState;
	public var energy:Int;

	public static var MAX_ENERGY:Int = 60 * 20;
	private static var SHOOTING_RATE:Int = 70;
	private static var MAX_SHOOTING_RATE:Int = 40;
	private static var START_SHOT_TIME:Int = -30;
	private var MIN_X:Int;
	private var MAX_X:Int;
	private var activeFrame:Int;
	private var healthBar:HealthBar;
	private var engine:Engine;
	private var smoke:EngineSmoking;
	private var movementPhase:MovementPhase;
	private var movementStep:Int;
	private var descending:Int;
	private var shooting:Bool;
	private var shots:Int;
	private var lastShot:Int;
	private var delayBetweenShots:Int;
	private var bulletPerShot:Int;
	private var bulletSpeeds:Array<Float>;
	private var tile_base:String;
	public var currLayer:String;

	public function new(parent:Stage, shipId:PlayerShip) {
		super(parent, "player");
		this.shipId = shipId;
		data = "player";

		tile_id = -1;
		tile_name = "player-" + (shipId == SHIP_A ? "a" : "b") + "-front";
		tile_base = "player-" + (shipId == SHIP_A ? "a" : "b");
		has_preupdate = true;
		radius = 4;

		descending = 0;
		health = 3;
		state = ACTIVE;
		energy = MAX_ENERGY;
		activeFrame = 0;
		currLayer = "front";

		bulletPerShot = 1;
		shooting = false;
		shots = 0;
		lastShot = START_SHOT_TIME;
		delayBetweenShots = 7;

		bulletSpeeds = [0.004, 0.002, 0.035];

		maxSpeed = 0.02;

		if (shipId == SHIP_A) {
			var posX:Int = 10;
			var posY:Int = 155;
			setCoordinates(posX, posY);

			nextMovementPhase(PHASE_1);
		} else {
			symetricFrom = Main.shipA;
		}

		healthBar = new HealthBar(parent, this);
		engine = new Engine(parent, this);
		smoke = new EngineSmoking(parent, this);
		smoke.alpha = 0;
	}

	private function nextMovementPhase(nextPhase:MovementPhase):Void {
		//dx = dy = 0;
		movementPhase = nextPhase;
		movementStep = 0;
		nextMovementStep(this);
	}

	private function nextMovementStep(e:Entity):Void {
		movementStep++;
		var destX:Float = x;
		var destY:Float = y;
		var MAX_Y:Int = Main.GAME_HEIGHT - 12;

		switch (movementPhase) {
			case PHASE_1:
			if (Main.nships == 2) {
				if (movementStep > 4) return nextMovementPhase(PHASE_2);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 4); destY = MAX_Y - 35;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 2) - 3; destY = MAX_Y - 15;

					case 3:
					destX = Std.int(Main.GAME_WIDTH / 4); destY = MAX_Y;

					case 4:
					destX = 3; destY = MAX_Y - 15;
				}
			} else {
				if (movementStep == 1 && x > Main.GAME_WIDTH / 2) return nextMovementPhase(PHASE_2);
				if (movementStep > 4) return nextMovementPhase(PHASE_2);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 3); destY = MAX_Y - 25;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 2); destY = MAX_Y - 5;

					case 3:
					destX = Std.int(2 * Main.GAME_WIDTH / 3); destY = MAX_Y - 25;

					case 4:
					destX = Main.GAME_WIDTH - 3; destY = MAX_Y - 35;
				}
			}

			case PHASE_2:
			if (Main.nships == 2) {
				if (movementStep > 2) return nextMovementPhase(PHASE_3);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 2) - 3; destY = MAX_Y - 35;

					case 2:
					destX = 3; destY = MAX_Y;
				}
			} else {
				if (movementStep == 1 && x < Main.GAME_WIDTH / 2) return nextMovementPhase(PHASE_3);
				if (movementStep > 6) return nextMovementPhase(PHASE_3);
				switch (movementStep) {
					case 1:
					destX = Main.GAME_WIDTH - 20; destY = MAX_Y - 10;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 3); destY = MAX_Y - 20;

					case 3:
					destX = Std.int(Main.GAME_WIDTH / 2); destY = MAX_Y - 5;

					case 4:
					destX = Std.int(2 * Main.GAME_WIDTH / 3); destY = MAX_Y - 20;

					case 5:
					destX = 20; destY = MAX_Y - 10;

					case 6:
					destX = 3; destY = MAX_Y - 35;
				}
			}

			case PHASE_3:
			if (Main.nships == 2) {
				if (movementStep > 3) return nextMovementPhase(PHASE_4);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 2) - 3; destY = MAX_Y;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 2) - 3; destY = MAX_Y - 35;

					case 3:
					destX = 3; destY = MAX_Y - 35;
				}
			} else {
				if (movementStep == 1 && x > Main.GAME_WIDTH / 2) return nextMovementPhase(PHASE_4);
				if (movementStep > 4) return nextMovementPhase(PHASE_4);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 3); destY = MAX_Y - 25;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 2); destY = MAX_Y - 5;

					case 3:
					destX = Std.int(2 * Main.GAME_WIDTH / 3); destY = MAX_Y - 25;

					case 4:
					destX = Main.GAME_WIDTH - 3; destY = MAX_Y - 35;
				}
			}

			case PHASE_4:
			if (Main.nships == 2) {
				if (movementStep > 2) return nextMovementPhase(PHASE_1);
				switch (movementStep) {
					case 1:
					destX = Std.int(Main.GAME_WIDTH / 2) - 3; destY = MAX_Y - 35;

					case 2:
					destX = 3; destY = MAX_Y - 15;
				}
			} else {
				if (movementStep == 1 && x < Main.GAME_WIDTH / 2) return nextMovementPhase(PHASE_1);
				if (movementStep > 6) return nextMovementPhase(PHASE_1);
				switch (movementStep) {
					case 1:
					destX = Main.GAME_WIDTH - 20; destY = MAX_Y - 10;

					case 2:
					destX = Std.int(Main.GAME_WIDTH / 3); destY = MAX_Y - 20;

					case 3:
					destX = Std.int(Main.GAME_WIDTH / 2); destY = MAX_Y - 5;

					case 4:
					destX = Std.int(2 * Main.GAME_WIDTH / 3); destY = MAX_Y - 20;

					case 5:
					destX = 20; destY = MAX_Y - 10;

					case 6:
					destX = 3; destY = MAX_Y - 35;
				}
			}
		}

		setDestination(destX, destY, nextMovementStep);
	}

	override private inline function preUpdate():Void {
		// Shoot if state == ACTIVE
		if (state == ACTIVE) {
			if (shooting) {
				if (activeFrame - lastShot >= delayBetweenShots) {
					shots++;
					shoot();
					if (shots >= 3) shooting = false;
				}

			} else {
				var shootingRate = (
					Main.nships == 2 &&
					Main.shipA.state == ACTIVE &&
					Main.shipB.state == ACTIVE
				) ? MAX_SHOOTING_RATE : SHOOTING_RATE;

				if (activeFrame - lastShot >= shootingRate) {
					shooting = true;
					shots = 1;
					shoot();
				}
			}
		} else {
			shooting = false;
			shots = 0;
		}

		energy += state == ACTIVE ? 1 : -2;
		energy = Std.int(Math.min(energy, MAX_ENERGY));
		if (energy < 0) {
			destroy();
		}
		activeFrame++;

		// Update movement speed
		maxSpeed = 0.02 + Math.min(Main.waveGen.diff, 8) * 0.002;

		// Update current tile
		var tileName:String = tile_base;

		// Update tile name according to current layer (= size)
		var currLayerDO:DisplayObject = _parent.player;
		if (currLayer == "mid") currLayerDO = _parent.midway;
		else if (currLayer == "safe") currLayerDO = _parent.safe;

		if (descending == 1) {
			descending++;
			engine.alpha = 0;
			_parent.switchLayer(this, currLayerDO, _parent.midway);
			tileName += "-mid";
			currLayer = "mid";
		} else if (descending == 2) {
			descending = 0;
			smoke.alpha = 1;
			tileName += "-safe";
			currLayer = "safe";
			_parent.switchLayer(this, currLayerDO, _parent.safe);
		} else if (descending == -1) {
			descending--;
			engine.alpha = 1;
			smoke.alpha = 0;
			tileName += "-mid";
			currLayer = "mid";
			_parent.switchLayer(this, currLayerDO, _parent.midway);
		} else if (descending == -2) {
			descending = 0;
			tileName += "-front";
			currLayer = "front";
			_parent.switchLayer(this, currLayerDO, _parent.player);
		} else if (state == ACTIVE) {
			tileName += "-front";
			currLayer = "front";
		} else {
			tileName += "-safe";
			currLayer = "safe";
		}

		// Update tile name according to current movement
		if (shipId == SHIP_A || Main.nships == 1) {
			if (dx > 0) tileName += "-right";
			else if (dx < 0) tileName += "-left";
		} else {
			if (Main.shipA.dx < 0) tileName += "-right";
			else if (Main.shipA.dx > 0) tileName += "-left";
		}

		if (tileName != tile_name) {
			tile_id = -1;
			tile_name = tileName;
		}
	}

	public function setState(newstate:PlayerState) {
		if (newstate == state) return;

		state = newstate;
		if (newstate == ACTIVE && activeFrame - lastShot >= SHOOTING_RATE) lastShot = activeFrame - SHOOTING_RATE + START_SHOT_TIME;

		if (newstate == PROTECTED) descending = 1;
		else descending = -1;

		Entity.needsUpdate = true;
	}

	public function powerUp(id:Int):Void {
		switch (id) {
			case 1:
			bulletSpeeds = [0.005, 0.0025, 0.045];
			delayBetweenShots = 5;

			case 2:
			SHOOTING_RATE = 55;
			MAX_SHOOTING_RATE = 30;

			case 3:
			bulletPerShot = 2;

			default:
		}
	}

	private function shoot():Void {
		if (Main.hasLost) return;

		// Display effect
		var shot:Shot = new Shot(_parent, "player", 2);
		shot.setCoordinates(x, y);

		var y:Int = this.y - 8;

		if (Main.nships == 2) {
			var distPlayers:Float = Math.abs(Main.shipA.x - Main.shipB.x) + Math.abs(Main.shipA.y - Main.shipB.y);
			if (Main.shipB.state == ACTIVE && distPlayers < 9 && Main.shipA.dy < 0) {
				if (dy == 0) return;

				var bullet:MaxiBullet = new MaxiBullet(_parent);
				bullet.setCoordinates(Main.GAME_WIDTH / 2, y);

				Main.shipA.shooting = false;
				Main.shipB.shooting = false;
				Main.shipA.lastShot = activeFrame + MAX_SHOOTING_RATE;
				Main.shipB.lastShot = activeFrame + MAX_SHOOTING_RATE;
				return;
			}
		}

		if (bulletPerShot == 1) {
			var bullet:Bullet = new Bullet(_parent, "player", shipId == SHIP_A ? "a" : "b", bulletSpeeds);
			bullet.setCoordinates(x, y);

		} else if (bulletPerShot == 2) {
			var bullet1:Bullet = new Bullet(_parent, "player", shipId == SHIP_A ? "a" : "b", bulletSpeeds);
			bullet1.setCoordinates(x - 2, y);

			var bullet2:Bullet = new Bullet(_parent, "player", shipId == SHIP_A ? "a" : "b", bulletSpeeds);
			bullet2.setCoordinates(x + 2, y);
		}

		lastShot = activeFrame;
	}

	override public function destroy():Void {
		if (isDeadEntity) return;

		healthBar.destroy();
		engine.destroy();
		smoke.destroy();

		Main.nships--;

		if (shipId == SHIP_A && Main.shipB.symetricFrom != null) {
			Main.shipB.symetricFrom = null;
			Main.shipB.movementPhase = movementPhase;
			Main.shipB.movementStep = movementStep;
			Main.shipB.setDestination(Main.GAME_WIDTH - destX, destY, Main.shipB.nextMovementStep);
		}

		if (!Main.hasLost) {
			var delay:Int = 0;
			for (i in 0...5) {
				var explosion:Explosion = new Explosion(_parent, delay);
				explosion.setCoordinates(
					x + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius)),
					y + Main.random.intBetween(Std.int(-1 * radius), Std.int(radius))
				);

				delay += 3;
			}
		}

		if (Main.nships <= 0) Main.gameover();
		super.destroy();
	}

	override public function dealDamage(amount:Int, angle:Int):Void {
		super.dealDamage(amount, angle);
	}

}
