package k.ld34.entities;

import k.lib.graphics.DisplayObject;
import k.lib.sound.SoundManager;
import k.lib.utils.KMath;
import k.lib.utils.Ease;

import k.ld34.entities.Entity;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

class Ship extends Entity {
	private var frame:Int;
	private var shootingRate:Int;
	public var minposx:Int;
	public var maxposx:Int;
	public var maxposy:Int;
	public var points:Int;

	private var hasHit:Bool;
	private var hasMove:Bool;
	private var baseTile:String;

	public function new(parent:Stage, ?diff:Int, ?x:Int, ?y:Int) {
		super(parent, "foes");

		tile_id = -1;
		tile_name = "ship";
		baseTile = "ship";

		points = 0;
		radius = 16;
		health = 1;
		frame = 1;
		has_preupdate = true;
		shootingRate = 90;
		minposx = maxposx = -1;
		maxposy = Main.GAME_HEIGHT + 20;

		hasHit = hasMove = false;
	}

	override private inline function preUpdate():Void {
		if (minposx != -1 && x < minposx) return destroy();
		if (maxposx != -1 && x > maxposx) return destroy();
		if (maxposy != -1 && y > maxposy) return destroy();

		_preUpdate();

		if (y > 0 && shootingRate > 0 && frame % shootingRate == 0) {
			shoot();
		}

		if (hasMove) {
			var tileName:String = baseTile;
			if (dx > 0) {
				tileName += "-right";
			} else if (dx < 0) {
				tileName += "-left";
			} else {
				tileName += "-base";
			}

			if (tileName != tile_name) {
				tile_id = -1;
				tile_name = tileName;
			}
		}

		frame++;
	}

	override public function dealDamage(amount:Int, angle:Int):Void {
		if (isDeadEntity) return;
		if (y < 0 || x < -10 || x > Main.GAME_WIDTH + 10) return;

		super.dealDamage(amount, angle);
	}

	override public function destroy():Void {
		if (health <= 0) Main.score += points;
		super.destroy();
	}

	private function _preUpdate():Void {}

	private function shoot():Void {
		if (Main.hasLost) return;
		if (isDeadEntity) return;
		if (y < 0 || x < -10 || x > Main.GAME_WIDTH + 10) return;

		var bulletSpeeds:Array<Float> = [0.001, 0.002, 0.015];
		var bullet = new Bullet(_parent, "foes", "", bulletSpeeds);
		bullet.setCoordinates(x, y);
	}

}
