package k.ld34.entities;

import k.lib.graphics.DisplayObject;
import k.lib.utils.KMath;
import k.lib.utils.Ease;
import k.lib.sound.SoundManager;

import k.ld34.effects.Shot;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

class ShipB extends Ship {
	private var shooting:Bool;
	private var shots:Int;
	private var lastShot:Int;
	private var dir:Int;
	private var lines:Int;

	private static var SHOOTING_RATE:Int = 65;
	private static var START_SHOT_TIME:Int = 10;

	public function new(parent:Stage, diff:Int, dir:Int) {
		super(parent);
		this.dir = dir;

		tile_id = -1;
		tile_name = "ship-b-base";
		points = 250;

		shooting = false;
		shots = 0;
		lines = 0;
		lastShot = START_SHOT_TIME;
		shootingRate = 0; // Disable default shooting

		radius = 10;
		health = 5;
		maxSpeed = 0.015;
		frame = Main.random.intBetween(0, 5);

		var x:Int = dir == 1 ? -20 : Main.GAME_WIDTH + 20;
		var y:Int = Main.random.intBetween(10, 40);
		setCoordinates(x, y);
		setDestination(Main.GAME_WIDTH - x, y, finishedLine);
	}

	private function finishedLine(e:Entity):Void {
		lines++;
		if (lines > 2) return destroy();

		if (lines % 2 == 1) {
			var y:Int = y + Main.random.intBetween(15, 30);
			setDestination(x, y, finishedLine);
		} else {
			var x:Int = dir == 1 ? -20 : Main.GAME_WIDTH + 20;
			setDestination(x, y, finishedLine);
			dir = -1 * dir;
		}
	}

	override private function _preUpdate():Void {
		if (shooting) {
			if (frame - lastShot >= 24) {
				shots++;
				shoot();
				if (shots >= 2) {
					shooting = false;
					lastShot += Main.random.intBetween(-10, 10);
				}
			}

		} else if (frame - lastShot >= SHOOTING_RATE) {
			shooting = true;
			shots = 1;
			shoot();
		}
	}

	override private function shoot():Void {
		if (Main.hasLost) return;

		// Display effect
		var shot:Shot = new Shot(_parent, "ship-b", 2);
		shot.setCoordinates(x, y);

		var bulletSpeeds:Array<Float> = [0.006, 0.003, 0.025];
		var bullet = new Bullet(_parent, "foes", "vertical", bulletSpeeds);
		bullet.setCoordinates(x, y + 10);
		lastShot = frame;
	}

}
