package k.ld34.entities;

import k.lib.graphics.DisplayObject;
import k.lib.utils.KMath;
import k.lib.utils.Ease;
import k.lib.sound.SoundManager;

import k.ld34.effects.Shot;

#if DEBUG
import k.lib.gconsole.GameConsole;
#end

class ShipE extends Ship {
	private var shooting:Bool;
	private var diff:Int;
	private var shots:Int;
	private var lastShot:Int;
	private var moves:Int;
	private var baseAngle:Float;

	private static var SHOOTING_RATE:Int = 200;
	private static var START_SHOT_TIME:Int = -150;
	private static var BULLETS_PER_SHOT:Int = 16;
	private static var N_SHOTS:Int = 2;

	public function new(parent:Stage, diff:Int) {
		super(parent);
		this.diff = diff;

		tile_id = -1;
		tile_name = "ship-c-base";
		points = 750;

		shooting = false;
		shots = 0;
		moves = 0;
		lastShot = START_SHOT_TIME;
		shootingRate = 0; // Disable default shooting
		baseAngle = 0;

		BULLETS_PER_SHOT = 12;
		if (diff > 2) BULLETS_PER_SHOT++;
		if (diff > 4) BULLETS_PER_SHOT++;
		if (diff > 6) BULLETS_PER_SHOT++;
		if (diff > 8) BULLETS_PER_SHOT++;

		N_SHOTS = 2;
		if (diff > 4) N_SHOTS++;
		if (diff > 8) N_SHOTS++;

		radius = 10;
		health = 15 + Std.int(Math.min(diff, 8));
		maxSpeed = 0.035;
		frame = Main.random.intBetween(0, 5);

		var x:Int = Main.random.intBelow(10) > 5 ? 10 : Main.GAME_WIDTH - 10;
		setCoordinates(x, -20);
		nextMove();
	}

	private function stopMoving(e:Entity):Void {
		moves++;
		dx = dy = 0;
		if (moves > 3) return destroy();
		shooting = true;
		shots = 0;
		lastShot = frame;
	}

	private function nextMove():Void {
		if (moves == 3) {
			var x:Int = Main.random.intBelow(10) > 5 ? -20 : Main.GAME_WIDTH + 20;
			setDestination(x, y, stopMoving);
		} else {
			var destX:Int = Main.random.intBetween(10, Main.GAME_WIDTH - 10);
			var destY:Int = Main.random.intBetween(10, 40);
			setDestination(destX, destY, stopMoving);
		}
	}

	override private function _preUpdate():Void {
		if (shooting) {
			if (frame - lastShot >= 24 - Math.min(6, diff/3)) {
				shots++;
				shoot();
				if (shots >= N_SHOTS) {
					shooting = false;
					lastShot += Main.random.intBetween(-10, 10);
				}
			}
		}


		if (frame % 360 == 0) {
			nextMove();
		}
	}

	override private function shoot():Void {
		if (Main.hasLost) return;

		// Display effect
		var shot:Shot = new Shot(_parent, "ship-c", 2);
		shot.setCoordinates(x, y);

		var angle:Float = baseAngle;
		var angleStep:Float = 6.28 / BULLETS_PER_SHOT;
		var posRadius:Int = 16;
		var pan:Float = 0.8 * (Main.GAME_WIDTH/2 - x) / Main.GAME_WIDTH;

		for (i in 0...BULLETS_PER_SHOT) {
			new ExpendingBullet(_parent, this, angle, posRadius, 0.5);
			angle += angleStep;
		}

		lastShot = frame;
		baseAngle += angleStep / 2;
	}

}
